/*
 **
 ** Server.java - Send sensor data using HTTP embedded server
 **
 ** Copyright (C) 2019 Elphel, Inc.
 **
 ** -----------------------------------------------------------------------------**
 **
 **  Server.java is free software: you can redistribute it and/or modify
 **  it under the terms of the GNU General Public License as published by
 **  the Free Software Foundation, either version 3 of the License, or
 **  (at your option) any later version.
 **
 **  This program is distributed in the hope that it will be useful,
 **  but WITHOUT ANY WARRANTY; without even the implied warranty of
 **  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 **  GNU General Public License for more details.
 **
 **  You should have received a copy of the GNU General Public License
 **  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ** -----------------------------------------------------------------------------**
 **
 */

package com.elphel.sensorserver;

import java.io.BufferedReader;
import java.io.IOException;
//import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.net.ServerSocket;
import java.net.Socket;
import java.nio.channels.IllegalBlockingModeException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.StringTokenizer;

import android.annotation.SuppressLint;
import android.app.Activity;
import android.util.Log;



public class Server {
	  private static String  TAG = "ServerSocket";
	  private static int     PORT = 8123;
	  private ServerSocket   server = null;
	  private Runnable       thconn;
//	  private final          SensorsUpdate sensorsUpdate;
	  public Server (
			  final Activity activity,
			  final SensorsUpdate sensorsUpdate
			  ) {
//		  this.sensorsUpdate = sensorsUpdate;
		  thconn = new Runnable() {
			  @Override
			public void run() {
				  try {
					  server = new ServerSocket(PORT);
				  } catch (IOException e1) {
					  // TODO Auto-generated catch block
					  Log.d(TAG,"Failed to create ServerSocket on port " + PORT+", " + e1.toString());
					  return;
				  }
				  while (true) {
					  Socket clientSocket = null;

					  try {
						  clientSocket = server.accept();
						  BufferedReader reader = new BufferedReader(new InputStreamReader(clientSocket.getInputStream()));
						  ArrayList<String> header = getHeader(reader);
						  Log.d(TAG,"header:\n" + header.toString());
						  HashMap<String,String> qmap = parseRequest(header);
						  if (qmap != null) {
							  Log.d(TAG,"qmap:\n" + qmap.toString());
						  } else {
							  Log.d(TAG,"Header parsing failed");
						  }
					  } catch (IOException e) {
						  Log.d(TAG,"IOException:" + e.getMessage());
						  return;
					  } catch (SecurityException e) {
						  Log.d(TAG,"SecurityException:" + e.getMessage());
						  return;
					  } catch (IllegalBlockingModeException e) {
						  Log.d(TAG,"IllegalBlockingModeException:" + e.getMessage());
						  return;
					  } catch (Exception e) {
						  Log.d(TAG,"Exception:" + e.toString());
						  return;
					  }

					  Log.d(TAG,"Request over");

					  float [] orientationAngles = sensorsUpdate.getOrientation();
					  double [] location = sensorsUpdate.getLocation();

					  float [] accelerometerReading  = sensorsUpdate.getAccelerometerReading();
					  float [] magnetometerReading =   sensorsUpdate.getMagnetometerReading();
					  long timestamp =                 sensorsUpdate.getTimeStamp();


//					  Date today = new Date();
					  String httpResponse = "HTTP/1.1 200 OK\r\n\r\n" + // today+"\r\n"+
							  orientationDegreeXml(timestamp, accelerometerReading, magnetometerReading, orientationAngles, location);
//					  String.format("<p>azimuth=%f</p>\r\n<p>pitch=%f</p>\r\n<p>roll=%f</p>\r\n",
//							  orientationAngles[0],orientationAngles[1],orientationAngles[2]);
					  try {
						  clientSocket.getOutputStream().write(httpResponse.getBytes("UTF-8"));
						  Log.i(TAG,"Response sent:"+httpResponse.replace("\r"," ").replace("\n"," "));
					  } catch (UnsupportedEncodingException e) {
						  Log.d(TAG,"UnsupportedEncodingException:" + e.toString());
					  } catch (IOException e) {
						  Log.d(TAG,"IOException while sending request:" + e.toString());
					  }
					  try {
						clientSocket.close();
					} catch (IOException e) {
						Log.d(TAG,"IOException while closing clientSocket:" + e.toString());
					}
//					  Log.d(TAG,"Response done:"+httpResponse);
				  }
			  }
		  };
	  }

	  /**
	   * Reads request header from the socket
	   * @param reader BufferedReader of the socket
	   * @return AddayList<String> of the header lines
	   * @throws IOException
	   */
	  private ArrayList<String> getHeader (BufferedReader reader) throws IOException {
		  ArrayList <String> header = new ArrayList<String>();
		  String line = reader.readLine();
		  while (!line.isEmpty()) {
			  header.add(line);
			  line = reader.readLine();
		  }
		  return header;
	  }

	  /**
	   * Parses request querry string fromthe header of the GET request
	   * @param header - request header as an ArrayList<String> of the header lines
	   * @return HashMap of strings:
	   * key == "?" - everything before the "?", starting with "/"
	   * other keys are passed parameters, value == "" if there is no "="
	   */
	  private HashMap<String,String> parseRequest(ArrayList<String> header){
		  HashMap<String,String> qmap = new HashMap<String,String>();
		  for (String line:header) {
			  StringTokenizer tokens = new StringTokenizer(line, " ");
			  if ((tokens.countTokens() >=3 ) &&  (tokens.nextToken().equals("GET"))){
				  String querry = tokens.nextToken();
				  StringTokenizer qparts = new StringTokenizer(querry, "?");
				  qmap.put("?", qparts.nextToken());
				  if (qparts.hasMoreTokens()) {
					  String qstr = qparts.nextToken();
					  StringTokenizer pairs = new StringTokenizer(qstr, "&");
					  for (int i =  pairs.countTokens(); i > 0; i--) {
						  StringTokenizer kv = new StringTokenizer(pairs.nextToken(), "=");
						  if (kv.countTokens()>0) {
							  String k = kv.nextToken();
							  String v = "";
							  if (kv.hasMoreTokens()) v = kv.nextToken();
							  qmap.put(k, v);
						  }
					  }
				  }
				  return qmap;
			  }
		  }
		  return null;
	  }


	  public void start() {
		  if (server == null) {
			  new Thread(thconn).start();
		  }
	  }
	  public void stop() {
		  if (server != null) {
			  try {
				  server.close();
				  server = null;
			  } catch (IOException e) {
				  e.printStackTrace();
			  }
		  }
	  }
	  private String orientationDegreeXml(float [] orientationAngles, double[] location) {
		  float [] oad = new float[orientationAngles.length];
		  for (int i = 0; i < oad.length; i++) {
			  oad[i] = orientationAngles[i]*180/ (float)Math.PI;
		  }
		  return orientationXml(oad, location);
	  }
	  private String orientationDegreeXml(long timestamp, float [] accelerometerReading, float [] magnetometerReading, float [] orientationAngles, double[] location) {
		  float [] oad = new float[orientationAngles.length];
		  for (int i = 0; i < oad.length; i++) {
			  oad[i] = orientationAngles[i]*180/ (float)Math.PI;
		  }
		  return orientationXml(timestamp, accelerometerReading, magnetometerReading, oad, location);
	  }
	  @SuppressLint("DefaultLocale")
	private String orientationXml(float [] orientationAngles, double[] location) {
		  return String.format(
				  "<?xml version=\"1.0\"?>\r\n"+
				  "<sensors>"+
				  "<azimuth>%f</azimuth>\r\n"+
				  "<pitch>%f</pitch>\r\n"+
				  "<roll>%f</roll>\r\n"+
				  "<longitude>%f</longitude>\r\n"+
				  "<latitude>%f</latitude>\r\n"+
				  "<date>%s</date>\r\n"+
				  "</sensors>",
				  orientationAngles[0],orientationAngles[1],orientationAngles[2], location[0], location[1], new Date());
	  }
	  @SuppressLint("DefaultLocale")
	private String orientationXml(long timestamp, float [] accelerometerReading, float [] magnetometerReading, float [] orientationAngles, double[] location) {
		  return String.format(
				  "<?xml version=\"1.0\"?>\r\n"+
				  "<sensors>"+
			      "<timestmap>%d</timestmap>\r\n"+
			      "<accelerometer>%f, %f, %f</accelerometer>\r\n"+
			      "<magnetometer>%f, %f, %f</magnetometer>\r\n"+
				  "<azimuth>%f</azimuth>\r\n"+
				  "<pitch>%f</pitch>\r\n"+
				  "<roll>%f</roll>\r\n"+
				  "<longitude>%f</longitude>\r\n"+
				  "<latitude>%f</latitude>\r\n"+
				  "<date>%s</date>\r\n"+
				  "</sensors>",
				  timestamp,
				  accelerometerReading[0], accelerometerReading[1], accelerometerReading[2],
				  magnetometerReading[0],  magnetometerReading[1],  magnetometerReading[2],
				  orientationAngles[0],orientationAngles[1],orientationAngles[2], location[0], location[1], new Date());
	  }


	  public int getPort() {
		  return PORT;
	  }

/*
private SensorManager sensorManager;
...
// Rotation matrix based on current readings from accelerometer and magnetometer.
final float[] rotationMatrix = new float[9];
SensorManager.getRotationMatrix(rotationMatrix, null,
    accelerometerReading, magnetometerReading);

// Express the updated rotation matrix as three orientation angles.
final float[] orientationAngles = new float[3];
SensorManager.getOrientation(rotationMatrix, orientationAngles);
 */

}
